from airflow.www.app import csrf
from dash import Dash
from flask_admin import expose, AdminIndexView
from flask import Markup
from flask import request



class DashAdminView(AdminIndexView, Dash):

    def __init__(self, external_stylesheets=None, **kwargs):
        AdminIndexView.__init__(self, **kwargs)
        url_base_of_dash = kwargs.get('url') + '/'
        Dash.__init__(self, url_base_pathname=url_base_of_dash, external_stylesheets=external_stylesheets, **kwargs)


    @expose('/')
    def index(self):
        """
        AdminView
        """

        return Dash.index(self)


    def _add_url(self, name, view_func, methods=('GET',)):
        """
        From Dash
        """
        csrf.exempt(view_func)
        self._urls.append((name.replace('/dash/', '/'), view_func.__name__, methods))
        self.routes.append(name)

    def interpolate_index(self,
                          metas='', title='', css='', config='',
                          scripts='', app_entry='', favicon='', renderer=''):
        """
        From Dash : interpolate_index
        """
        content = self.render('brian/dash.html',
                              app_entry=app_entry,
                              dash_config=config,
                              scripts=scripts)
        return '{}'.format(content)



    @expose('/version')
    def version(self):
        content = "Plugin's Version : {}"
        import pkg_resources
        version = pkg_resources.get_distribution("ddui").version

        return self.render('brian/dash.html', app_entry=content.format(version), title="Datadriver - Version")


# Graph dash core components needs to be imported here (module level) to work fine
from dash_core_components import Graph, Dropdown
from dash_dagre import DagreD3
from dash_html_components import Div, H5
from dash_table import DataTable
from plotly.graph_objs import Histogram, Figure, Layout


def dags_dashtable(dags, html_id):

    dags_list = Dropdown(
        id=html_id,
        options=[
            {'label': dag_id, 'value': dag_id} for dag_id in dags
        ],
    )

    return dags_list


def col_histograms(df):
    """
    Generate plotly.Dash's Bar chart component. One barchart for each column of the dataframe
    :param df: pandas.DataFrame
    :return: dash_core_components.Graph
    """
    describe_bar_chart = Div(children=[])
    for col in df.columns:
        t = Histogram(x=df[col], name=col)
        l = Layout(title='Histogram for colunm {}'.format(t.name),
            autosize=False,
            width=500,
            height=500,
            margin={
                'l': 50,
                'r': 50,
                'b': 100,
                't': 100,
                'pad': 4
            })
        fig = Figure(data=[t], layout=l)
        g = Graph(figure=fig,className="col-sm-4", config={'staticPlot': True, 'responsive': True})
        describe_bar_chart.children.append(g)

    return describe_bar_chart


def describe_dashtable(df):
    """
    Generate a dash_table.DataTable component with data from dataframe.describe() function
    :param df:
    :return:
    """
    describe_dataframe = df.describe(include='all').reset_index()
    describe_datatable = Div(className="table-responsive-sm-6", children=[
        H5("Pandas' describe"),
        DataTable(
            columns=[{"name": i, "id": i} for i in describe_dataframe.columns],
            data=describe_dataframe.to_dict("rows")
        )])
    return describe_datatable


def head_and_tail_tables(df):
    head = df.head()
    tail = df.tail()
    head_and_tail = Div(children=[
        Div(className="table-responsive-sm-4", children=[
            H5("Pandas' Head"),
            DataTable(
                columns=[{"name": i, "id": i} for i in head.columns],
                data=head.to_dict("rows")
            )]),
        Div(className="table-responsive-sm-4", children=[
            H5("Pandas' Tail"),
            DataTable(
                columns=[{"name": i, "id": i} for i in tail.columns],
                data=tail.to_dict("rows")
            )])
        ,
    ])
    return head_and_tail


def dashdag(dag=None, html_id=None):
    """
    Generate a DagreD3 component, showing a Directed Acyclic Graph from Airflow's models.DagBag object
    :param dag_roots:
    :param tasks:
    :return:
    """
    if dag is None :
        return Div(DagreD3(id=html_id, nodes=[], edges=[]))

    task_id_mapping = {}
    nodes = {}
    edges = []
    for i, task in enumerate(dag.tasks):
        task_id_mapping[task.task_id] = i
        nodes[i] = {
            'parent_dag_id': dag.dag_id,
            'label': task.task_id,
            'labelStyle': "fill:{0};".format(task.ui_fgcolor),
            'style': "fill:{0};".format(task.ui_color),
            'rx': 5,
            'ry': 5,
            'output_table': getattr(task, "output_table", "")
        }

    def get_upstream(task):
        for t in task.upstream_list:
            edge = [task_id_mapping[t.task_id], task_id_mapping[task.task_id], {'curve': 'd3.curveBasis'}]

            if edge not in edges:
                edges.append(edge)
                get_upstream(t)

    for t in dag.roots:
        get_upstream(t)

    return Div(DagreD3(id=html_id, nodes=nodes, edges=edges, width='50%'), style={"background-color": " #f2f2f2"})



import os

import click

from pkg_resources import resource_filename

@click.group()
def cli():
    pass

@cli.command()
def install():
    from airflow.configuration import get
    install_dir = os.path.dirname(resource_filename("ddui", "__init__.py"))

    root_airflow_plugin = get('core', 'plugins_folder')
    if not os.path.exists(root_airflow_plugin):
        os.mkdir(root_airflow_plugin)
        click.echo('Create Airflow plugins folder {} '.format(root_airflow_plugin))
    plugins_folder = os.path.join(root_airflow_plugin, 'ddui')

    if not os.path.exists(plugins_folder):
        os.symlink(install_dir, plugins_folder)
        click.echo('Add {} to Airflow\'s plugins folder in {}'.format(install_dir, plugins_folder))
    else:
        click.echo('''Plugin {} is already installed.
                Run `python setup.py clean` if you want to cleanup and restart the process'''.format(plugins_folder))

@cli.command()
def clean():
    from airflow.configuration import conf
    plugins_folder = os.path.join(conf.get('core', 'plugins_folder'), 'ddui')
    if os.path.exists(plugins_folder):
        os.remove(plugins_folder)
        click.echo('Remove plugins folder {} from Airflow\'s plugins'.format(plugins_folder))
    else:
        click.echo('No link to remove : {} does not exist'.format(plugins_folder))



from dash_html_components import Div, Button, Span, Strong


def Panel(head, body, idHtml=None, className=None):
    return Div(id=idHtml, className=f"panel panel-default {className}", children=[
        Div(head, className="panel-heading"),
        Div(body, className="panel-body")
    ])


def Alert(message):
    return Div(id="alert-div",
               className="alert alert-warning alert-dismissible",
               children=[
                   Button(className="close", type='button',
                          children=[Span("x")], **{"data-dismiss": "alert", "aria-label": "Close"}),
                   Strong("Warning!"),
                   f" {message}",
               ]
               )


def Error(message):
    return Div(id="alert-div",
               className="alert alert-danger alert-dismissible",
               children=[
                   Strong("Error!"),
                   f" {message}",
               ]
               )



import logging
from copy import copy

from dash.dependencies import Output, Input
from dash_core_components import Tab, Tabs
from dash_html_components import Div
from flask_admin.consts import ICON_TYPE_IMAGE

from ddui import orm
# Graph dash core components needs to be imported here (because it will be included as a dash's javascript component)
from ddui import plot
from ddui.dash_components import Panel, Alert, Error
from ddui.views import DashAdminView

TABLE_OF_DAGS_HTML_ID = 'table-dags'
DAG_VIEW_HTML_ID = 'div-dag'
DAGRE_HTML_ID = 'dagre-component'
PLOT_DESCRIPTION_HTML_ID = 'describe-div'

app = DashAdminView(category='DataDriver',
                    name='Datascience UI',
                    endpoint='dash',
                    url='/dash',
                    menu_icon_type=ICON_TYPE_IMAGE,
                    menu_icon_value='brian/sigle.png',
                    external_stylesheets=["https://stackpath.bootstrapcdn.com/bootstrap/4.2.1/css/bootstrap.min.css"])

app.scripts.config.serve_locally = False
app.css.config.serve_locally = False
app.config['suppress_callback_exceptions'] = True

dags_df = orm.airflow_dags()

app.layout = Div(className='container', id='ds-ui', children=[
    Div(className="row", children=[
        Panel(head="Select a DAG", body=plot.dags_dashtable(dags_df, TABLE_OF_DAGS_HTML_ID), idHtml="div-select-dag",
              className="col-md-3"),
        Panel(head="", body="", className="col-md-12", idHtml=DAG_VIEW_HTML_ID),
    ]),
    Div(className="row", id=PLOT_DESCRIPTION_HTML_ID),
])


@app.callback(Output(DAG_VIEW_HTML_ID, 'children'),
              [Input(TABLE_OF_DAGS_HTML_ID, "value")])
def show_dag(selected_dag):
    from airflow import models, settings

    dagbag = models.DagBag(settings.DAGS_FOLDER)

    dag = dagbag.get_dag(selected_dag)

    return [Div('Click on a task to see its output', className="panel-heading"),
            Div(plot.dashdag(dag, html_id=DAGRE_HTML_ID), className="panel-body")]


@app.callback(Output(PLOT_DESCRIPTION_HTML_ID, 'children'),
              [Input(DAGRE_HTML_ID, "selectedId"),
               Input(DAGRE_HTML_ID, "selectedNodeProps")])
def show_columns(selectedId, selectedNodeProps):
    if not selectedId or not selectedNodeProps:
        return Alert('Select a task to watch the output !')

    output_table = selectedNodeProps.get('output_table', '')
    if not output_table:
        return Alert('This operator has no output_table property. Is it a DataDriver workflow ?')
    dag_id = selectedNodeProps.get('parent_dag_id', '')

    from airflow import models, settings

    dagbag = models.DagBag(settings.DAGS_FOLDER)
    fileloc = dagbag.get_dag(dag_id).fileloc
    db = _get_db_from_datadriver_dag(fileloc)
    if db is None:
        return Alert("Object named 'db' not found in {}".format(fileloc))

    logging.info("DB type is {} \t\t {}".format(type(db), db))
    try:
        df = db.retrieve_table(output_table)
    except Exception as e:
        return Error(f"An exception occured when reading output_table {output_table} : {e}")

    describe_datatable = plot.describe_dashtable(df)
    describe_bar_chart = plot.col_histograms(df)
    head_and_tail_tables = plot.head_and_tail_tables(df)
    tab_head_tail = Tab(label='Head and tail', children=[head_and_tail_tables    ])
    tab_bar_chart = Tab(label='Bar-chart by column', children=[describe_bar_chart])
    tab_describe = Tab(label='Describe frame', children=[describe_datatable])

    return Panel(
        head="Select a Tab to see the statistics on the task's output",
        body=Tabs([
            tab_describe,
            tab_bar_chart,
            tab_head_tail
        ], colors={"border": "white", "primary": "#00a9c5", "background": "#e0e0e0"})
    )


def _get_db_from_datadriver_dag(filepath):
    with open(filepath) as f:
        logging.info("Trying to get DB object from {}".format(filepath))
        local_dict = {}
        exec(f.read(), local_dict)

        if 'db' in local_dict:
            return local_dict.get('db')
        else:
            objects = copy(local_dict)
            for obj in objects.values():
                if hasattr(obj, 'retrieve_table'):
                    return obj





def airflow_dags():
    """
    Read the list of dags from Airflow's metastore DB and return the list as a pandas.DataFrame
    :return: pandas.DataFrame
    """
    from airflow import models, settings

    dagbag = models.DagBag(settings.DAGS_FOLDER)

    return list(dagbag.dags)


from flask import Blueprint
from flask_admin.base import MenuLink
from flask_admin.consts import ICON_TYPE_IMAGE

from ddui.dash_app import app as dash_view
from airflow.plugins_manager import AirflowPlugin


ml_repo_website = MenuLink(
    category='DataDriver',
    name='Git repository',
    url='https://gitlab.octo.com/dd/ddui.git',
    icon_type=ICON_TYPE_IMAGE,
    icon_value='brian/git.png'
)

ml_doc = MenuLink(
    category='DataDriver',
    name='DataDriver API documentation',
    url='http://datadriver-doc-ddapi.s3-website-eu-west-1.amazonaws.com/',
    icon_type=ICON_TYPE_IMAGE,
    icon_value='brian/sigle.png'
)

ml_version = MenuLink(
    category='DataDriver',
    name='Version',
    url='/dash/version',
    icon_type=ICON_TYPE_IMAGE,
    icon_value='brian/sigle.png'
)


brian_bp = Blueprint(
    "brian_web", __name__,
    template_folder='templates',
    static_folder='static/brian',
    static_url_path='/static/brian',
)

class DataDriverUIPlugin(AirflowPlugin):
    name = 'DataDriver UI Plugin'
    operators = []
    hooks = []
    executors = []
    macros = []
    admin_views = [dash_view]
    flask_blueprints = [brian_bp]
    menu_links = [ml_doc, ml_repo_website, ml_version]